䕓䕌呃吠偏ㄠ〰‰‪剆䵏圠湯ㅖ弶佃呎䅒呃਍佇                                                                               
 Usage: pisql [OPTIONS] COMMAND [ARGS]...                                      
                                                                               
 A simple but rich command line interface for Sybase ASE                       
                                                                               
+- Options -------------------------------------------------------------------+
| --install-completion          Install completion for the current shell.     |
| --show-completion             Show completion for the current shell, to     |
|                               copy it or customize the installation.        |
| --help                        Show this message and exit.                   |
+-----------------------------------------------------------------------------+
+- Commands ------------------------------------------------------------------+
| dev      
